/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "driver.hpp"

Driver::Driver(Wordlist &wl, Settings &se, InputReader &ir, Damerau &da, SplitCompound &ss, IntroducerList &il, SoundError &serr)
     : wordlist(wl)
     , settings(se) 
     , inputreader(ir)
     , damerau(da)
     , splitCompound(ss)
     , introducerlist(il)
     , sounderror(serr)
     , activeDamerau(settings.getBool("damerau/active", true))
     , activeSplitCompound(settings.getBool("splitCompound/active", true))
     , activeSoundError(settings.getBool("soundError/active", true))
     , sentenceSeparatorTag(settings.getString("options/sentenceSeparatorTag", "mad"))
     , maxErrorsInSentence(settings.getInt("options/maxErrorsInSentence", 32768))     
     , generator(settings.getInt("options/seed", 123456789))     
     , random_bit_damerau(generator, distribution_type((double)settings.getDouble("damerau/probability", 10) / 100))
     , random_bit_splitCompound(generator, distribution_type((double)settings.getDouble("splitCompound/probability", 10) / 100))
     , random_bit_sounderror(generator, distribution_type((double)settings.getDouble("soundError/probability", 10) / 100)) {
    
    for (std::list<std::pair<ErrorIntroducer, double> >::const_iterator iter = introducerlist.getList().begin(); iter != introducerlist.getList().end(); ++iter) {
        random_bits_introducer.push_back(std::make_pair(bernoulli(generator, distribution_type(iter->second)), 0));
    }
}

void Driver::createErrors() {
     int numErrorsInSentence = 0;
     int semSoundError = 0;
     int semSplitCompound = 0;
     int semDamerau = 0;
     try {
	  while (true) {
	       if (wordlist.tag(inputreader.getBuffer().front().tag()) == sentenceSeparatorTag || inputreader.getBuffer().front().blank()) {
		    //std::cerr << "New sentence. " << numErrorsInSentence << " errors in last sentence\n";
		    std::cerr << ".";
		    numErrorsInSentence = 0;
	       }
	       if (numErrorsInSentence < maxErrorsInSentence) {
		    std::list<std::pair<bernoulli,int> >::iterator bit_iter = random_bits_introducer.begin();
		    for (std::list<std::pair<ErrorIntroducer, double> >::const_iterator iter = introducerlist.getList().begin(); iter != introducerlist.getList().end(); ++iter) {
			 if (bit_iter->first()) {
			      ++(bit_iter->second);
			 }
			 if (bit_iter->second > 0 && iter->first.apply(inputreader.getBuffer())) {
			      --(bit_iter->second);
			      ++numErrorsInSentence;
			 }
			 ++bit_iter;
		    }
		    if (activeSoundError) {
			 if (random_bit_sounderror()) {
			      ++semSoundError;
			 }
			 if (semSoundError > 0 && sounderror.apply(inputreader.getBuffer())) {
			      --semSoundError;
			      ++numErrorsInSentence;
			 }
		    }
		    if (activeSplitCompound) {
			 if (random_bit_splitCompound()) {
			      ++semSplitCompound;
			 }
			 if (semSplitCompound > 0) {
			      bool ididit = false;
			      // FIXME kontrollera att fromIndex och toIndex blir korrekta
			      // vid multipel srskrivning
			      while (splitCompound.apply(inputreader.getBuffer())) {
				   ididit = true;
				   ++numErrorsInSentence;
			      }
			      if (ididit) {
				   --semSplitCompound;
			      }
			 }			
		    }
		    if (activeDamerau) {
			 if (random_bit_damerau()) {
			      ++semDamerau;
			 }
			 if (semDamerau > 0 && 
			     damerau.apply(inputreader.getBuffer())) {
			      --semDamerau;
			      ++numErrorsInSentence;
			 }
		    }
	       }
	       inputreader.next();
	  }
     }
     catch (EndOfInputException &) {
	  std::cout << "Done!\n";
     }
     catch (...) {
	  std::cerr << "Unexpected exception\n";
     }
     std::cout << std::endl;
}
