/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#ifndef ERRORINTRODUCER_H
#define ERRORINTRODUCER_H

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include <list>

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/PlatformUtils.hpp>

#include <boost/tokenizer.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_io.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>

#include "utils.hpp"
#include "settings.hpp"
#include "wordtag.hpp"
#include "wordlist.hpp"
#include "exceptions.hpp"
#include "errorwriter.hpp"
#include "myerrorhandler.hpp"
#include "minedit.hpp"
#include "inputreader.hpp"

XERCES_CPP_NAMESPACE_USE

typedef boost::tuple<std::string, std::string, std::string> RuleTuple;

struct RuleOp : public std::binary_function<RuleTuple, RuleTuple, bool> {
     const std::string alphabet;
     RuleOp() : alphabet("abcdefghijklmnopqrstuvwxyz") { }
     bool operator()(RuleTuple rt1, RuleTuple rt2) const {
	  std::string s1 = rt1.get<0>().substr(0, rt1.get<0>().find_first_not_of(alphabet));
	  std::string s2 = rt2.get<0>().substr(0, rt2.get<0>().find_first_not_of(alphabet));
	  return s1 < s2;
     }
};

class ErrorIntroducer {
public:

     ErrorIntroducer(Wordlist &, Settings &, ErrorWriter &, InputReader &, const std::string &);
     void printRules() const;
     bool apply(std::list<WordTag>&) const;
     bool applyHelper(const std::string&, const std::string&, const std::string&, std::list<WordTag>&) const;
private:

     typedef boost::char_separator<char> Separator;
     typedef boost::tokenizer<Separator> Tokenizer;

     typedef std::vector<RuleTuple> RuleVector;
     typedef std::multiset<RuleTuple, RuleOp> RuleSet;

     typedef boost::tuple<WordTag, WordTag, bool> WtWtBTuple;
     typedef std::list<WtWtBTuple> TupleList;
     typedef std::vector<WtWtBTuple> TupleVector;

     void readFromFile(const std::string&);
     std::string replaceMatchedExpressions(const std::string&, const boost::match_results<std::string::const_iterator>&) const;
     std::string reorder(const std::string&, const std::string&) const;
     std::list<WordTag> reorder(const std::list<WordTag>&, const std::string&) const;
     void newOrder(const std::string&, const std::string&, const std::string&, std::list<WordTag>&) const;
     void printList(const TupleList &) const;
     void printVector(const TupleVector &) const;
     Wordlist &wordlist;
     Settings &settings;
     ErrorWriter &errorwriter;
     InputReader &inputreader;
     const char escapeChar;
     std::string spaceChar;
     const char wordChar;
     RuleSet sortedRules;
     RuleVector unsortedRules;
     bool allowExistingWords;
     bool forceAllowWords;
     bool allowTagChange;
     bool forceAllowTag;
     std::string reportErrorIntroducer;
};

#endif
