/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "inputreader.hpp"

InputReader::InputReader(Wordlist &wl, Settings &se)
     : wordlist(wl)
     , settings(se)
     , infile(settings.getString("input/filename", "input.wt").c_str())
     , outfile(settings.getString("output/filename", "output.wt").c_str(), std::ios::out|std::ios::trunc)
     , format(settings.getString("output/format", "%1%\t%2%"))
     , bufferSize(settings.getInt("input/bufferSize", 5))
     , row_read(0)
     , row_orig(0)
     , noError(settings.getString("output/description/noError", "ok"))
     , existingWord(settings.getString("output/description/existingWord", "exist"))
     , nonExistingWord(settings.getString("output/description/nonExistingWord", "noexist"))
     , wordChange(settings.getString("output/description/wordChange", "-wordch"))
     , noWordChange(settings.getString("output/description/noWordChange", "-nowordch"))
     , tagChange(settings.getString("output/description/tagChange", "-tagch"))
     , noTagChange(settings.getString("output/description/noTagChange", "-notagch")) {     
     if (!infile) {
	  std::cerr << "Error: Could not open input file\n";
	  std::exit(1);
     }
     if (!outfile) {
	  std::cerr << "Error: Could not open output file\n";
	  std::exit(1);
     }
     exp.set_expression(settings.getString("input/expression", "(.+)\\t(.+)\\t.*"), boost::regbase::normal);
     try {
	  for (unsigned int i = 0; i < bufferSize; i++) {
	       buffer.push_back(readRow());
	  }
     }
     catch (EndOfInputException &) {
	  std::cerr << "Warning: End of input already while buffering.\n";
     }
     catch (InputParseException &) {
	  std::cerr << "Error: Input parse exception.\n";
     }
     catch (...) {
	  std::cerr << "Error: Unexpected exception ir1.\n";
     }
}

InputReader::~InputReader() {

}

void InputReader::next() {
     try {
	  if (buffer.size() <= bufferSize) {
	       buffer.push_back(readRow());
	  }
	  else {
	       //std::cerr << "Buffer has length " << buffer.size() << ", no read needed.\n";
	  }
     }
     catch (EndOfInputException &) {
	  //std::cerr << "Warning: End of input.\n";
     }
     catch (InputParseException &) {
	  std::cerr << "Error: Input parse exception.\n";
     }
     catch (...) {
	  std::cerr << "Error: Unexpected exception ir2.\n";
     }

     try {
	  const WordTag &iter = buffer.front();
	  if (iter.blank()) {
	       if (iter.rest()[0] == '\r') {
		    outfile << std::endl;
	       }
	       else {
		    outfile << iter.rest() << std::endl;
	       }
	  }
	  else {
	       std::string lemma;
	       try {
		    lemma = wordlist.findLemma(iter.word(), iter.tag());
	       }
	       catch (NotFoundException &) {
		    lemma = wordlist.unknownLemma();
	       }
	       std::string description = iter.desc();
	       if (description == "") {
		    description = noError;
	       }
	       else {
		    description += "/";
		    description += wordlist.lookup(iter.word()) ? existingWord : nonExistingWord;
		    description += (iter.word() == iter.origWord()) ? noWordChange : wordChange;
		    description += (iter.tag() == iter.origTag()) ? noTagChange : tagChange;
	       }
	       boost::format my_format(format);
	       my_format.exceptions(boost::io::all_error_bits ^ boost::io::too_many_args_bit);
	       outfile << my_format % iter.word() % wordlist.tag(iter.tag()) % lemma % iter.rest() % description << std::endl;
	  }
     }
     catch (boost::io::format_error &e) {
	  std::cerr << "Format error in output/format: \n" << e.what() << std::endl;
     }

     buffer.pop_front();
     if (buffer.empty()) {
	  throw EndOfInputException();
     }
}

WordTag InputReader::readRow() {
     boost::match_results<std::string::const_iterator> result;
     std::string line;
     if (!std::getline(infile, line)) {
	  throw EndOfInputException();
     }
     if (!regex_match(line, result, exp)) {
	  //throw InputParseException();
	  return WordTag("", 1, ++row_read, "", 1, ++row_orig, line, true);
     }
     return WordTag(result[1], wordlist.untag(result[2]), ++row_read, result[1], wordlist.untag(result[2]), ++row_orig, result[3], false);
}

void InputReader::printBuffer() const {
     for (std::list<WordTag>::const_iterator iter = buffer.begin();
	  iter != buffer.end(); ++iter) {
	  std::cout << "Word: " << iter->word() << " \tTag: " << iter->tag() << "\n";
     }
}

std::list<WordTag>& InputReader::getBuffer() {
     return buffer;
}

void InputReader::changeReadOffset(int num) {
     row_read += num;
}
