/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "sounderror.hpp"

SoundError::SoundError(Wordlist &wl, Settings &se, ErrorWriter &ew)
     : wordlist(wl)
     , settings(se)
     , errorwriter(ew)
     , escapeChar(settings.getChar("options/escapeChar", '\\'))
     , generator(settings.getInt("options/seed", 123456789))
     , allowExistingWords(settings.getBool("soundError/allowExistingWords", true))
     , forceAllowWords(settings.getBool("soundError/forceAllowWords", false))
     , allowTagChange(settings.getBool("soundError/allowTagChange", true))
     , forceAllowTag(settings.getBool("soundError/forceAllowTag", false)) 
     , reportSoundError(settings.getString("soundError/reportName", "sound")) {
     std::string configDir = settings.getString("options/configDir", "");
     readFromFile(configDir + settings.getString("soundError/filename", "sound.test"));
}

void SoundError::readFromFile(const std::string& filename) {
     std::ifstream infile(filename.c_str());
     if (!infile) {
	  std::cerr << "Error: Could not open file '" << filename << "'\n";
	  return;
     }
     boost::match_results<std::string::const_iterator> result;
     boost::regex exp;
     exp.set_expression(settings.getString("soundError/expression", "(.+)\\t(.+)\\t(.+)"), boost::regbase::normal);
     std::string line;
     try {
	  // For each row in the input file
	  while (std::getline(infile, line)) {
	       if (regex_match(line, result, exp)) {
		    if (result.size() != 4) {
			 throw InputParseException("Wrong number of tokens read.");
		    }
		    // Insert item
		    std::string num = result[3];
		    insert(result[1], result[2], boost::lexical_cast<double>(num));
	       }
	  }
     }
     catch (const boost::bad_lexical_cast &e) {
	  std::cerr << "Error: " << e.what() << std::endl;
     }
     catch (const InputParseException &e) {
	  std::cerr << "Error: " << e.what() << std::endl;
     }
}

void SoundError::insert(const std::string &match, const std::string &to, double slh) {
     boost::regex exp(match, boost::regbase::normal);
     SoundError::Datamap::iterator iter = data.find(exp);
     // If other items have been compiled to the same regular expression
     if (iter != data.end()) {
	  // Update probability sum for these expressions
	  iter->second.first += slh;
	  // Add the expression to transform to to the list
	  iter->second.second.push_back(std::make_pair(to, slh));
     }
     // Else insert a new expression
     else {
	  // Create a new list for this expression
	  SoundError::ListPStrDbl slhlist;
	  slhlist.push_back(std::make_pair(to, slh));
	  data.insert(std::make_pair(exp, std::make_pair(slh, slhlist)));
     }
}

bool SoundError::apply(std::list<WordTag>& listwt) const {
     if (listwt.front().blank()) {
	  return false;
     }
     boost::match_results<std::string::const_iterator> result;
     WordTag &wt = listwt.front();
     
     Datamap newdata;
     double newdataSum = 0;
     for (SoundError::Datamap::const_iterator iter = data.begin(); iter != data.end(); ++iter) {
	  if (regex_match(wt.word(), result, iter->first)) {
	       newdata.insert(*iter);
	       newdataSum += iter->second.first;
	       //std::cerr << "Match:" << iter->first.str() << " matched " << wt.word() << ".\n";
	  }
     }
     if (newdataSum == 0) {
	  return false;
     }
     random_real random_rule(generator, distribution_type(0.0, newdataSum));
     double rule = random_rule();
     SoundError::Datamap::const_iterator rule_iter = newdata.begin();
     double current_rule = rule_iter->second.first;
     while (rule > current_rule && rule_iter != newdata.end()) {
	  ++rule_iter;
	  if (rule_iter != newdata.end()) {
	       current_rule += rule_iter->second.first;
	  }
     }
     if (rule_iter != newdata.end()) {
	  //std::cerr << "using rule " << rule_iter->first.str() << "\n";

	  if (regex_match(wt.word(), result, rule_iter->first)) {
	       // Create a random number between 0 and the sum of this list
	       random_real random_number(generator, distribution_type(0.0, rule_iter->second.first));
	       double number = random_number();
	       SoundError::ListPStrDbl::const_iterator subiter = rule_iter->second.second.begin();
	       double current = subiter->second;
	       while (number > current && subiter != rule_iter->second.second.end()) {
		    ++subiter;
		    if (subiter != rule_iter->second.second.end()) {
			 current += subiter->second;
		    }
	       }
	       if (subiter != rule_iter->second.second.end()) {
		    WordTag wtnew = wt;
		    std::string newword = replaceMatchedExpressions(subiter->first, result);
		    // Check if a change was done
		    if (newword != listwt.front().word()) {
			 wtnew.set(newword, wordlist.getTag(newword, wt.tag()));
	 
			 // Allow only if word exists and existing words are allowed,
			 // or if word doesn't exist and existing words are not obligatory.
			 if (wordlist.lookup(newword) && allowExistingWords ||
			     !wordlist.lookup(newword) && !(allowExistingWords && forceAllowWords)) {
			      //std::cerr << "Word allowed. ";
			      // Allow only if tag has changed and tag change is allowed,
			      // or if tag didn't change and tag change is not obligatory
			      if (wtnew.tag() != listwt.front().tag() && allowTagChange ||
				  wtnew.tag() == listwt.front().tag() && !(allowTagChange && forceAllowTag)) {
				   //std::cerr << "Tag allowed.\n";
				   wtnew.addDesc(reportSoundError);
				   wtnew.setWord(wordlist.preserveCase(listwt.front().word(), wtnew.word()));
				   errorwriter.insertSoundError(wt, wtnew);
				   listwt.pop_front();
				   listwt.push_front(wtnew);
				   return true;
			      }
			 }
		    }
	       }
	  }

     }
     return false;
}

std::string SoundError::replaceMatchedExpressions(const std::string& exp, const boost::match_results<std::string::const_iterator> &result) const {
     std::string ret = exp;
     std::string::size_type index = ret.find(escapeChar, 0);
     try {
	  while (index != std::string::npos) {
	       unsigned int param = boost::lexical_cast<unsigned int>(ret.at(index + 1));
	       if (param >= 1 && param <= result.size() - 1) {
		    ret.replace(index, 2, result[param]);
	       }
	       else {
		    throw std::out_of_range("Invalid back reference");
	       }
	       index = ret.find(escapeChar, index);
	  }
     }
     catch (const boost::bad_lexical_cast &) {
	  std::cerr << "Error: " << escapeChar << " must be followed by a number\n";
     }
     catch (const std::out_of_range &e) {
	  std::cerr << "Error: out of range: " << e.what() << "\n";
     }
     return ret;
}
